<?php

namespace UnicaenDbAnonym\Service;

use Doctrine\DBAL\ConnectionException;
use Doctrine\DBAL\Exception;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\ClassMetadata;
use Faker\Factory;
use Laminas\Log\Logger;
use Laminas\Log\Writer\Stream;
use Locale;
use RuntimeException;
use Throwable;
use UnicaenApp\Service\SQL\RunSQLProcess;
use UnicaenApp\Service\SQL\RunSQLResult;
use Webmozart\Assert\Assert;

class DbService
{
    const METADATA_KEY_ANONYMISEE = 'BDD_ANONYMISEE';

    /**
     * @var \Doctrine\ORM\EntityManager
     */
    protected $entityManager;

    /**
     * @var \Doctrine\DBAL\Platforms\AbstractPlatform
     */
    protected $databasePlatform;

    /**
     * @var string
     */
    protected $updateSQLTemplate = 'update %s set %s where id = %d ;';

    /**
     * @var \Faker\Generator $faker
     */
    protected $faker;


    public function __construct()
    {
        $this->faker = Factory::create(Locale::getDefault());
    }

    /**
     * @param \Doctrine\ORM\EntityManager $entityManager
     * @throws \Doctrine\DBAL\Exception
     */
    public function setEntityManager(EntityManager $entityManager): void
    {
        $this->entityManager = $entityManager;
        $this->databasePlatform = $this->entityManager->getConnection()->getDatabasePlatform();
    }

    public function isBddAnonymisee(): bool
    {
        $sqlTemplate = "select value from _METADATA where key = '%s'";

        $sql = sprintf($sqlTemplate, self::METADATA_KEY_ANONYMISEE);

        try {
            /** @var string|false $result */
            $result = $this->entityManager->getConnection()->executeQuery($sql)->fetchOne();
        } catch (Throwable $e) {
            throw new RuntimeException(
                "Erreur lors de l'interrogation du témoin d'anonymisation dans la table des metadata", null, $e);
        }

        if ($result === false) {
            throw new RuntimeException(sprintf(
                "Le témoin d'anonymisation '%s' est introuvable dans la table des metadata",
                self::METADATA_KEY_ANONYMISEE
            ));
        }

        return $result === '1';
    }

    public function setBddAnonymisee(bool $bddAnonymisee)
    {
        $sqlTemplate = "update _METADATA set value = '%d' where key = '%s'";

        $sql = sprintf($sqlTemplate, $bddAnonymisee ? '1' : '0', self::METADATA_KEY_ANONYMISEE);

        try {
            $this->entityManager->getConnection()->executeQuery($sql);
        } catch (Exception $e) {
            throw new RuntimeException(
                "Erreur lors de la modification du témoin d'anonymisation dans la table des metadata", null, $e);
        }
    }

    public function fetchEntityRecords(string $entityName, array $fields, array $except): array
    {
        $qb = $this->entityManager->createQueryBuilder()
            ->select("partial t.{id," . implode(',', $fields) . "}")
            ->from($entityName, 't');

        foreach ($except as $field => $values) {
            $qb->andWhere($qb->expr()->notIn('t.' . $field, (array)$values));
        }

        return $qb->getQuery()->getArrayResult();
    }

    public function getEntityClassMetadata(string $entityName): ClassMetadata
    {
        return $this->entityManager->getClassMetadata($entityName);
    }

    public function genUpdateForAnonymiser(array $record, array $mapping, ClassMetadata $metadata): string
    {
        $tableName = $metadata->getTableName();

        $sets = [];

        foreach ($mapping as $fieldName => $fakerConfig) {
            $fakerConfig = (array)$fakerConfig;
            $method = array_shift($fakerConfig);
            if ($method === 'null') {
                $fakeValue = 'NULL';
            } else {
                $fakeValue = $this->faker->$method(...$fakerConfig);
                $fakeValue = $this->databasePlatform->quoteStringLiteral($fakeValue);
            }
            $columnName = $metadata->getColumnName($fieldName);
            $sets[] = "$columnName = $fakeValue";
        }

        return sprintf($this->updateSQLTemplate,
            $tableName,
            implode(', ', $sets),
            $record['id']
        );
    }

    public function genUpdateForRestaurer(array $record, array $mapping, ClassMetadata $metadata): string
    {
        $tableName = $metadata->getTableName();

        $sets = [];

        foreach ($mapping as $fieldName => $fakerConfig) {
            $value = $this->databasePlatform->quoteStringLiteral($record[$fieldName]);
            $columnName = $metadata->getColumnName($fieldName);
            $sets[] = "$columnName = $value";
        }

        return sprintf($this->updateSQLTemplate,
            $tableName,
            implode(', ', $sets),
            $record['id']
        );
    }

    /**
     * @throws \Doctrine\DBAL\ConnectionException
     */
    public function lancerScript(string $scriptPath): RunSQLResult
    {
        Assert::readable($scriptPath, "Le script '$scriptPath' n'existe pas ou n'est pas lisible");

        $conn = $this->entityManager->getConnection();
        $conn->beginTransaction();

        $p = new RunSQLProcess();
        $p->setConnection($conn);
        $p->setScriptPath($scriptPath);
        $p->setQueriesSplitPattern("#;\n#m"); // point virgule puis un retour
        $p->setLogFilePath($scriptPath . '.log');
        $p->setLogger((new Logger())->addWriter(new Stream('php://stdout')));
        $result = $p->executeScript();

        try {
            $conn->commit();
        } catch (ConnectionException $e) {
            $conn->rollBack();
            throw $e;
        }

        return $result;
    }
}